﻿// Fill out your copyright notice in the Description page of Project Settings.


#include "GOAPAgent.h"

#include "Actions/Combo_P.h"


// Sets default values for this component's properties
UGOAPAgent::UGOAPAgent()
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;

	// ...
}


// Called when the game starts
void UGOAPAgent::BeginPlay()
{
	Super::BeginPlay();

	//Gets all classes that inherit from UGOAPAction and adds them to the AvailableActions array
	TArray<UClass*> ActionClasses;
	GetDerivedClasses(UGOAPAction::StaticClass(), ActionClasses, false);
	for (UClass* ActionClass : ActionClasses)
	{
		UGOAPAction* Action = NewObject<UGOAPAction>(GetWorld(), ActionClass);
		Action->Init();
		AvailableActions.Add(Action);
	}

	//Sort AvailableActions by cost
	AvailableActions.Sort([](const UGOAPAction& A, const UGOAPAction& B) { return A.ActionCost < B.ActionCost; });
}

bool UGOAPAgent::BuildActionGraph(UActionNode* Parent, TArray<UActionNode*>& SuccessfulLeaves, UWorldState* WorldState, TMap<FString, int> AgentGoals)
{
	bool bFoundAPlan = false;

	for (UGOAPAction* Action : AvailableActions)
	{
		if (Action->CheckPreConditions(WorldState))
		{
			UWorldState* NewWorldState = Action->ApplyEffects(WorldState);
			UActionNode* NewNode = NewObject<UActionNode>();
			NewNode->Init(Parent, Action, NewWorldState, Parent->Cost + Action->ActionCost);

			if (NewWorldState->State["PlayerHealth"] <= 0)
			{
				SuccessfulLeaves.Add(NewNode);
				bFoundAPlan = true;
			}
			else if (!SeenStates.Contains(NewWorldState) && !bFoundAPlan)
			{
				SeenStates.Add(NewWorldState);
				TArray<UActionNode*> NewLeaves;
				if (BuildActionGraph(NewNode, NewLeaves, NewWorldState, AgentGoals))
				{
					SuccessfulLeaves.Append(NewLeaves);
					bFoundAPlan = true;
				}
			}
		}
	}

	//Log the plan
	if (bFoundAPlan)
	{
		UE_LOG(LogTemp, Warning, TEXT("Found a plan!"));
		// UActionNode* CurrentNode = SuccessfulLeaves[0];
		// while (CurrentNode->Parent != nullptr)
		// {
		// 	UE_LOG(LogTemp, Warning, TEXT("%s"), *CurrentNode->Action->GetName());
		// 	CurrentNode = CurrentNode->Parent;
		// }
	}
	return bFoundAPlan;
}

// Called every frame
void UGOAPAgent::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	// ...
}

TArray<UGOAPAction*> UGOAPAgent::Plan(UWorldState* WorldState, TMap<FString, int> AgentGoals)
{
	TArray<UGOAPAction*> ActionPlan;
	UActionNode* RootActionNode = NewObject<UActionNode>();
	RootActionNode->Init(nullptr, nullptr, WorldState, 0);
	TArray<UActionNode*> SuccessfulLeaves;

	SeenStates.Empty();
	SeenStates.Add(WorldState);

	if (BuildActionGraph(RootActionNode, SuccessfulLeaves, WorldState, AgentGoals))
	{
		//Find the cheapest leaf
		float LowestRunningCost = 9999999;
		UActionNode* CheapestLeaf = nullptr;
		for (UActionNode* Leaf : SuccessfulLeaves)
		{
			if (Leaf->Cost < LowestRunningCost)
			{
				LowestRunningCost = Leaf->Cost;
				CheapestLeaf = Leaf;
			}
		}

		//Get the action plan
		TArray<UActionNode*> Result;
		UActionNode* ActionNode = CheapestLeaf;
		while (ActionNode != nullptr)
		{
			Result.Add(ActionNode);
			ActionNode = ActionNode->Parent;
		}

		//Reverse the array and only get the actions
		for (int i = Result.Num() - 1; i >= 0; i--)
		{
			ActionPlan.Add(Result[i]->Action);
		}
	}

	//Remove the first action as it is null due to being the root node
	ActionPlan.RemoveAt(0);
	return ActionPlan;
}
